﻿using System;
using System.Collections.Generic;
using System.Threading;
using System.Threading.Tasks;
using HeuristicLab.Algorithms.PushGP.Expressions;
using HeuristicLab.Algorithms.PushGP.Generators;
using HeuristicLab.Algorithms.PushGP.Stack;

namespace HeuristicLab.Algorithms.PushGP.Interpreter
{
    public class PushGPInterpreter : IInterpreter
    {
        private Task currentTask;


        public PushGPInterpreter(Configuration config = null)
        {
            Configuration = config ?? new Configuration();

            CodeStack = new PushGPStack<Expression>(128);
            ExecStack = new PushGPStack<Expression>(128);
            NameStack = new PushGPStack<string>(16);
            BooleanStack = new PushGPStack<bool>(16);
            IntegerStack = new PushGPStack<long>(16);
            FloatStack = new PushGPStack<double>(16);

            CustomExpressions = new Dictionary<string, Expression>();

            CodeGenerator = new CodeGenerator(this);
        }

        public Configuration Configuration { get; private set; }

        public int ExecCounter { get; private set; }
        public bool IsPaused { get; private set; }
        public bool IsAborted { get; private set; }
        public bool IsRunning { get { return ExecStack.Count > 0 && !IsPaused && !IsAborted && ExecCounter < Configuration.EvalPushLimit; } }
        public bool IsCompleted { get { return ExecStack.Count == 0; } }
        public bool CanStep { get { return !IsCompleted && !IsAborted && IsPaused; } }

        public IStack<Expression> CodeStack { get; private set; }
        public IStack<Expression> ExecStack { get; private set; }
        public IStack<string> NameStack { get; private set; }
        public IStack<bool> BooleanStack { get; private set; }
        public IStack<long> IntegerStack { get; private set; }
        public IStack<double> FloatStack { get; private set; }

        public IDictionary<string, Expression> CustomExpressions { get; private set; }

        public CodeGenerator CodeGenerator { get; private set; }

        public bool IsNameQuoteFlagSet { get; set; }

        public void Reset()
        {
            IsAborted = false;
            IsPaused = false;
            currentTask = null;

            Clear();
        }

        public void Clear()
        {
            ExecCounter = 0;

            ExecStack.Clear();
            CodeStack.Clear();
            NameStack.Clear();
            BooleanStack.Clear();
            IntegerStack.Clear();
            FloatStack.Clear();

            CustomExpressions.Clear();
        }

        public void Interpret(string code)
        {
            var program = Encode(code);
            Interpret(program);
        }

        public Task InterpretAsync(string code, bool paused = false, CancellationToken token = default(CancellationToken))
        {
            var program = Encode(code);

            currentTask = InterpretAsync(program, paused, token);

            return currentTask;
        }

        public Task InterpretAsync(Expression program, bool paused = false, CancellationToken token = default(CancellationToken))
        {
            IsPaused = paused;
            currentTask = Task.Run(() => Interpret(program), token);

            return currentTask;
        }

        public void Interpret(Expression program)
        {
            /* Push top expression so the loop is able to enter
             * If the top expression is a single statement then the loop has nothing to do
             * Otherwise the expand expression will be evaluated and pushes code onto the EXEC stack */
            ExecStack.Push(program);

            if (Configuration.TopLevelPushCode)
            {
                CodeStack.Insert(0, program);
            }

            // run top expression
            ExecStack.Pop().Eval(this);

            Interpret();
        }

        private void Interpret()
        {
            while (IsRunning)
            {
                DoStep();
                ExecCounter++;
            }

            Finally();
        }

        private void Finally()
        {
            if (IsCompleted && Configuration.TopLevelPopCode && CodeStack.Count > 0)
            {
                CodeStack.Pop();
            }
        }

        private void DoStep()
        {
            ExecStack.Pop().Eval(this);
        }

        private Task InterpreteAsync()
        {
            currentTask = Task.Run(() => Interpret());

            return currentTask;
        }

        public async Task AbortAsync()
        {
            if (IsAborted || IsCompleted)
                return;

            IsAborted = true;

            if (currentTask != null)
            {
                await currentTask;
            }
        }

        public async Task AbortAndResetAsync()
        {
            await AbortAsync();
            Reset();
        }

        public async Task PauseAsync()
        {
            if (IsPaused || IsCompleted)
                return;

            IsPaused = true;

            if (currentTask != null)
            {
                await currentTask;
            }
        }

        public Task ResumeAsync()
        {
            if (IsPaused || !IsAborted)
            {
                IsPaused = false;
                return InterpreteAsync();
            }

            return currentTask;
        }

        public void Step()
        {
            if (CanStep)
            {
                DoStep();
                Finally();
            }
        }

        public void Step(int count)
        {
            for (var i = 0; i < count; i++)
            {
                Step();
            }
        }

        public static Expression Encode(string code)
        {
            return Parser.Parse(code);
        }

        public static Task<Expression> EncodeAsync(string code, CancellationToken token = default(CancellationToken))
        {
            return Task.Run(() => Encode(code), token);
        }

        public static string Decode(Expression expression)
        {
            return string.Concat(expression.ToString());
        }

        public static Task<string> DecodeAsync(Expression expression, CancellationToken token = default(CancellationToken))
        {
            return Task.Run(() => Decode(expression), token);
        }

        public void PrintStacks()
        {
            PrintStack("EXEC", ExecStack);
            PrintStack("CODE", CodeStack);
            PrintStack("NAME", NameStack);
            PrintStack("BOOLEAN", BooleanStack);
            PrintStack("FLOAT", FloatStack);
            PrintStack("INTEGER", IntegerStack);
        }

        private void PrintStack<T>(string name, IStack<T> stack)
        {
            Console.WriteLine(string.Format(
                "--- {0} ---\n{1}\n---------------------------------------------------",
                name,
                stack));
        }
    }
}
